"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getTsconfig = void 0;
const utils_1 = require("@umijs/utils");
const path_1 = __importDefault(require("path"));
const typescript_transform_paths_1 = __importDefault(require("typescript-transform-paths"));
const constants_1 = require("../../../constants");
const utils_2 = require("../../../utils");
/**
 * get parsed tsconfig.json for specific path
 */
function getTsconfig(cwd) {
    // use require() rather than import(), to avoid jest runner to fail
    // ref: https://github.com/nodejs/node/issues/35889
    const ts = require('typescript');
    const tsconfigPath = ts.findConfigFile(cwd, ts.sys.fileExists);
    if (tsconfigPath) {
        const tsconfigFile = ts.readConfigFile(tsconfigPath, ts.sys.readFile);
        return ts.parseJsonConfigFileContent(tsconfigFile.config, ts.sys, path_1.default.dirname(tsconfigPath));
    }
}
exports.getTsconfig = getTsconfig;
/**
 * get declarations for specific files
 */
async function getDeclarations(inputFiles, opts) {
    const enableCache = process.env.FATHER_CACHE !== 'none';
    const tscCacheDir = path_1.default.join(opts.cwd, constants_1.CACHE_PATH, 'tsc');
    if (enableCache) {
        // make tsc cache dir
        utils_1.fsExtra.ensureDirSync(tscCacheDir);
    }
    const output = [];
    // use require() rather than import(), to avoid jest runner to fail
    // ref: https://github.com/nodejs/node/issues/35889
    const ts = require('typescript');
    const tsconfig = getTsconfig(opts.cwd);
    if (tsconfig) {
        // check tsconfig error
        /* istanbul ignore if -- @preserve */
        if (tsconfig.errors.length) {
            throw new Error(`Error parsing tsconfig.json content: ${utils_1.chalk.redBright(ts.flattenDiagnosticMessageText(tsconfig.errors[0].messageText, '\n'))}`);
        }
        // warn if noEmit is false
        /* istanbul ignore if -- @preserve */
        if (tsconfig.options.declaration && tsconfig.options.noEmit === true) {
            utils_2.logger.warn('tsconfig.json `noEmit` is true, will not emit declaration files!');
            return output;
        }
        // enable declarationMap by default in development mode
        if (process.env.NODE_ENV === 'development' &&
            tsconfig.options.declaration &&
            tsconfig.options.declarationMap !== false) {
            tsconfig.options.declarationMap = true;
        }
        // remove paths which out of cwd, to avoid transform to relative path by ts-paths-transformer
        Object.keys(tsconfig.options.paths || {}).forEach((item) => {
            const pathAbsTarget = path_1.default.resolve(tsconfig.options.pathsBasePath, tsconfig.options.paths[item][0]);
            if (!(0, utils_1.winPath)(pathAbsTarget).startsWith(`${(0, utils_1.winPath)(opts.cwd)}/`)) {
                delete tsconfig.options.paths[item];
                utils_2.logger.debug(`Remove ${item} from tsconfig.paths, because it's out of cwd.`);
            }
        });
        // enable incremental for cache
        if (enableCache && typeof tsconfig.options.incremental === 'undefined') {
            tsconfig.options.incremental = true;
            tsconfig.options.tsBuildInfoFile = path_1.default.join(tscCacheDir, 'tsconfig.tsbuildinfo');
        }
        const tsHost = ts.createIncrementalCompilerHost(tsconfig.options);
        tsHost.writeFile = (fileName, content, _a, _b, sourceFiles) => {
            const sourceFile = sourceFiles === null || sourceFiles === void 0 ? void 0 : sourceFiles[0].fileName;
            // only collect dts for input files, to avoid output error in watch mode
            // ref: https://github.com/umijs/father/issues/43
            if (sourceFile && inputFiles.includes(sourceFile)) {
                const ret = {
                    file: path_1.default.basename(fileName),
                    content,
                    sourceFile,
                };
                output.push(ret);
            }
            else if (fileName === tsconfig.options.tsBuildInfoFile) {
                utils_1.fsExtra.writeFileSync(tsconfig.options.tsBuildInfoFile, content);
            }
        };
        const incrProgram = ts.createIncrementalProgram({
            rootNames: inputFiles,
            options: tsconfig.options,
            host: tsHost,
        });
        // using ts-paths-transformer to transform tsconfig paths to relative path
        // reason: https://github.com/microsoft/TypeScript/issues/30952
        // ref: https://www.npmjs.com/package/typescript-transform-paths
        const result = incrProgram.emit(undefined, undefined, undefined, true, {
            afterDeclarations: [
                (0, typescript_transform_paths_1.default)(incrProgram.getProgram(), { afterDeclarations: true }, 
                // specific typescript instance, because this plugin is incompatible with typescript@4.9.x currently
                // but some project may declare typescript and some dependency manager will hoist project's typescript
                // rather than father's typescript for this plugin
                { ts }),
            ],
        });
        // check compile error
        // istanbul-ignore-if
        if (result.diagnostics.length) {
            result.diagnostics.forEach((d) => {
                const loc = ts.getLineAndCharacterOfPosition(d.file, d.start);
                const rltPath = (0, utils_1.winPath)(path_1.default.relative(opts.cwd, d.file.fileName));
                const errMsg = ts.flattenDiagnosticMessageText(d.messageText, '\n');
                utils_2.logger.error(`${utils_1.chalk.blueBright(rltPath)}:${
                // correct line number & column number, ref: https://github.com/microsoft/TypeScript/blob/93f2d2b9a1b2f8861b49d76bb5e58f6e9f2b56ee/src/compiler/tracing.ts#L185
                `${utils_1.chalk.yellow(loc.line + 1)}:${utils_1.chalk.yellow(loc.character + 1)}`} - ${utils_1.chalk.gray(`TS${d.code}:`)} ${errMsg}`);
            });
            throw new Error('Declaration generation failed.');
        }
    }
    return output;
}
exports.default = getDeclarations;
